"""Provides `Expr` and its subclasses."""

from __future__ import annotations

import ast
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    SupportsIndex,
    TypeVar,
    Union,
    cast,
    overload,
)

ConstType = Union[None, str, bytes, bool, int, float]
PassedType = TypeVar(
    "PassedType",
    bound=Union[ast.cmpop, ast.operator, ast.boolop, ast.unaryop, ast.expr_context],
)
T = TypeVar("T")
T2 = TypeVar("T2", bound=Union[ConstType, "Expr"])
V = TypeVar("V", bound=ConstType)

if TYPE_CHECKING:
    from collections.abc import Iterator, Mapping, Sequence
    from types import CodeType

    from pydantic.annotated_handlers import GetCoreSchemaHandler
    from pydantic_core import core_schema
    from typing_extensions import TypedDict, Unpack

    from ._context_keys import ContextKey

    # Used for node end positions in constructor keyword arguments
    _EndPositionT = TypeVar("_EndPositionT", int, None)

    # Corresponds to the names in the `_attributes`
    # class variable which is non-empty in certain AST nodes
    class _Attributes(TypedDict, Generic[_EndPositionT], total=False):
        lineno: int
        col_offset: int
        end_lineno: _EndPositionT
        end_col_offset: _EndPositionT


def parse_expression(expr: Expr | str) -> Expr:
    """Parse string expression into an [`Expr`][app_model.expressions.Expr] instance.

    Parameters
    ----------
    expr : Expr | str
        Expression to parse.  (If already an `Expr`, it is returned)

    Returns
    -------
    Expr
        Instance of `Expr`.

    Raises
    ------
    SyntaxError
        If the provided string is not an expression (e.g. it's a statement), or
        if it uses any forbidden syntax components (e.g. Call, Attribute,
        Containers, Indexing, Slicing, f-strings, named expression,
        comprehensions.)
    """
    if isinstance(expr, Expr):
        return expr
    try:
        # mode='eval' means the expr must consist of a single expression
        tree = ast.parse(str(expr), mode="eval")
        if not isinstance(tree, ast.Expression):
            raise SyntaxError  # pragma: no cover
        return ExprTransformer().visit(tree.body)
    except SyntaxError as e:
        raise SyntaxError(f"{expr!r} is not a valid expression: ({e}).") from None


def safe_eval(expr: str | bool | Expr, context: Mapping | None = None) -> Any:
    """Safely evaluate `expr` string given `context` dict.

    This lets you evaluate a string expression with broader expression
    support than `ast.literal_eval`, but much less support than `eval()`.
    It also supports booleans (which are returned directly), and `Expr` instances,
    which are evaluated in the given `context`.

    Parameters
    ----------
    expr : str | bool | Expr
        Expression to evaluate. If `expr` is a string, it is parsed into an
        `Expr` instance. If a `bool`, it is returned directly.
    context : Mapping | None
        Context (mapping of names to objects) to evaluate the expression in.
    """
    if isinstance(expr, bool):
        return expr
    return parse_expression(expr).eval(context)


class Expr(ast.AST, Generic[T]):
    """Base Expression class providing dunder and convenience methods.

    This is a subclass of `ast.AST` that provides rich dunder methods that
    facilitate joining and comparing typed expressions.  It only implements a
    subset of ast Expressions (for safety of evaluation), but provides more
    than `ast.literal_eval`.

    Expressions that are supported:

    - Names: 'myvar' (these must be evaluated along with some context)
    - Constants: '1'
    - Comparisons: 'myvar > 1'
    - Boolean Operators: 'myvar & yourvar' (bitwise `&` and `|` are overloaded
      here to mean boolean `and` and `or`)
    - Binary Operators: 'myvar + 42' (includes `//`, `@`, `^`)
    - Unary Operators: 'not myvar'

    Things that are *NOT* supported:

    - attribute access: 'my.attr'
    - calls: 'f(x)'
    - containers (lists, tuples, sets, dicts)
    - indexing or slicing
    - joined strings (f-strings)
    - named expressions (walrus operator)
    - comprehensions (list, set, dict, generator)
    - statements & assignments (e.g. 'a = b')

    This class is not meant to be instantiated directly. Instead, use
    [`parse_expression`][app_model.expressions._expressions.parse_expression], or the
    [`Expr.parse`][app_model.expressions.Expr.parse] classmethod to create an expression
    instance.

    Once created, an expression can be joined with other expressions, or
    constants.


    Examples
    --------
    >>> expr = parse_expression("myvar > 5")

    combine expressions with operators
    >>> new_expr = expr & parse_expression("v2")

    nice `repr`
    >>> new_expr
    BoolOp(
        op=And(),
        values=[
            Compare(
            left=Name(id='myvar', ctx=Load()),
            ops=[
                Gt()],
            comparators=[
                Constant(value=5)]),
            Name(id='v2', ctx=Load())])

    evaluate in some context
    >>> new_expr.eval(dict(v2="hello!", myvar=8))
    'hello!'

    you can also use keyword arguments.  This is *slightly* slower
    >>> new_expr.eval(v2="hello!", myvar=4)

    serialize
    >>> str(new_expr)
    'myvar > 5 and v2'

    One reason you might want to use this object is to capture named expressions
    that can be evaluated repeatedly as some underlying context changes.

    ```python
    light_is_green = Name[bool]("light_is_green")
    count = Name[int]("count")
    is_ready = light_is_green & count > 5

    assert is_ready.eval({"count": 4, "light_is_green": True}) == False
    assert is_ready.eval({"count": 7, "light_is_green": False}) == False
    assert is_ready.eval({"count": 7, "light_is_green": True}) == True
    ```

    this will also preserve type information:
    >>> reveal_type(is_ready())  # revealed type is `bool`
    """

    _names: set[str]
    _code: CodeType

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        if type(self).__name__ == "Expr":
            raise RuntimeError("Don't instantiate Expr. Use `Expr.parse`")
        super().__init__(*args, **kwargs)
        self._recompile()

    def _recompile(self) -> None:
        ast.fix_missing_locations(self)
        self._code = compile(ast.Expression(body=self), "<Expr>", "eval")  # type: ignore [arg-type]
        self._names = set(self._iter_names())

    def eval(
        self, context: Mapping[str, object] | None = None, **ctx_kwargs: object
    ) -> T:
        """Evaluate this expression with names in `context`."""
        if context is None:
            context = ctx_kwargs
        elif ctx_kwargs:
            context = {**context, **ctx_kwargs}
        try:
            return eval(self._code, {}, context)  # type: ignore
        except NameError as e:
            miss = {k for k in self._names if k not in context}
            raise NameError(
                f"Names required to eval this expression are missing: {miss}"
            ) from e

    @classmethod
    def parse(cls, expr: str) -> Expr:
        """Parse string into Expr (classmethod).

        see docstring of [`parse_expression`][app_model.expressions.parse_expression]
        for details.
        """
        return parse_expression(expr)

    def __str__(self) -> str:
        """Serialize this expression to string form."""
        return self._serialize()

    def _serialize(self) -> str:
        """Serialize this expression to string form."""
        return str(_ExprSerializer(self))

    def __repr__(self) -> str:
        return f"Expr.parse({str(self)!r})"

    @staticmethod
    def _cast(obj: Any) -> Expr:
        """Cast object into an Expression."""
        return obj if isinstance(obj, Expr) else Constant(obj)

    # boolean operators
    # '&' and '|' are normally binary operators... but we use them here to
    # combine expression objects meaning "and" and "or".
    # if you want the binary operators, use Expr.bitand, and Expr.bitor

    def __and__(
        self, other: Expr[T2] | Expr[T] | ConstType | Compare
    ) -> BoolOp[T | T2]:
        return BoolOp(ast.And(), [self, other])

    def __or__(self, other: Expr[T2] | Expr[T] | ConstType | Compare) -> BoolOp[T | T2]:
        return BoolOp(ast.Or(), [self, other])

    # comparisons

    def __lt__(self, other: Any) -> Compare:
        return Compare(self, [ast.Lt()], [other])

    def __le__(self, other: Any) -> Compare:
        return Compare(self, [ast.LtE()], [other])

    def __eq__(self, other: Any) -> Compare:  # type: ignore
        return Compare(self, [ast.Eq()], [other])

    def __ne__(self, other: Any) -> Compare:  # type: ignore
        return Compare(self, [ast.NotEq()], [other])

    def __gt__(self, other: Any) -> Compare:
        return Compare(self, [ast.Gt()], [other])

    def __ge__(self, other: Any) -> Compare:
        return Compare(self, [ast.GtE()], [other])

    # using __contains__ always returns a bool... so we provide our own
    # Expr.in_ and Expr.not_in methods

    def in_(self, other: Any) -> Compare:
        """Return a comparison for `self` in `other`."""
        # not a dunder, use with Expr.in_(a, other)
        return Compare(self, [ast.In()], [other])

    def not_in(self, other: Any) -> Compare:
        """Return a comparison for `self` no in `other`."""
        return Compare(self, [ast.NotIn()], [other])

    # binary operators
    # (note that __and__ and __or__ are reserved for boolean operators.)

    def __add__(self, other: T | Expr[T]) -> BinOp[T]:
        return BinOp(self, ast.Add(), other)

    def __sub__(self, other: T | Expr[T]) -> BinOp[T]:
        return BinOp(self, ast.Sub(), other)

    def __mul__(self, other: T | Expr[T]) -> BinOp[T]:
        return BinOp(self, ast.Mult(), other)

    def __truediv__(self, other: T | Expr[T]) -> BinOp[T]:
        return BinOp(self, ast.Div(), other)

    def __floordiv__(self, other: T | Expr[T]) -> BinOp[T]:
        return BinOp(self, ast.FloorDiv(), other)

    def __mod__(self, other: T | Expr[T]) -> BinOp[T]:
        return BinOp(self, ast.Mod(), other)

    def __matmul__(self, other: T | Expr[T]) -> BinOp[T]:
        return BinOp(self, ast.MatMult(), other)  # pragma: no cover

    def __pow__(self, other: T | Expr[T]) -> BinOp[T]:
        return BinOp(self, ast.Pow(), other)

    def __xor__(self, other: T | Expr[T]) -> BinOp[T]:
        return BinOp(self, ast.BitXor(), other)

    def bitand(self, other: T | Expr[T]) -> BinOp[T]:
        """Return bitwise self & other."""
        return BinOp(self, ast.BitAnd(), other)

    def bitor(self, other: T | Expr[T]) -> BinOp[T]:
        """Return bitwise self | other."""
        return BinOp(self, ast.BitOr(), other)

    # unary operators

    def __neg__(self) -> UnaryOp[T]:
        return UnaryOp(ast.USub(), self)

    def __pos__(self) -> UnaryOp[T]:
        # usually a no-op
        return UnaryOp(ast.UAdd(), self)

    def __invert__(self) -> UnaryOp[T]:
        # note: we're using the invert operator `~` to mean "not ___"
        return UnaryOp(ast.Not(), self)

    def __reduce_ex__(self, protocol: SupportsIndex) -> tuple[Any, ...]:
        rv: list[Any] = list(super().__reduce_ex__(protocol))
        rv[1] = tuple(getattr(self, f) for f in self._fields)
        return tuple(rv)

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source: type, handler: GetCoreSchemaHandler
    ) -> core_schema.CoreSchema:
        # this will only be called by pydantic v2
        from pydantic_core import core_schema

        return core_schema.no_info_plain_validator_function(cls._validate)

    @classmethod
    def _validate(cls, v: Any) -> Expr:
        """Validate v as an `Expr`. For use with Pydantic."""
        return v if isinstance(v, Expr) else parse_expression(v)

    def __hash__(self) -> int:
        _hash = hash(self.__class__)
        for f in self._fields:
            field = getattr(self, f)
            if isinstance(field, list):
                field = tuple(field)
            _hash += hash(field)
        return _hash

    def _iter_names(self) -> Iterator[str]:
        yield from _iter_names(self)


LOAD = ast.Load()


class Name(Expr[T], ast.Name):
    """A variable name.

    Parameters
    ----------
    id : str
        The name of the variable.
    bound : Any | None
        The type of the variable represented by this name (i.e. the type to which this
        name evaluates to when used in an expression). This is used to provide type
        hints when evaluating the expression. If `None`, the type is not known.
    """

    def __init__(
        self,
        id: str,
        ctx: ast.expr_context = LOAD,
        *,
        bound: type[T] | None = None,
        **kwargs: Unpack[_Attributes],
    ) -> None:
        super().__init__(id, ctx=ctx, **kwargs)
        self.bound = bound


class Constant(Expr[V], ast.Constant):
    """A constant value.

    Parameters
    ----------
    value : V
        the Python object this constant represents.
        Types supported: NoneType, str, bytes, bool, int, float
    kind : str | None
        The kind of constant.  This is used to provide type hints when
    """

    value: V  # pyright: ignore[reportIncompatibleVariableOverride]

    def __init__(
        self, value: V, kind: str | None = None, **kwargs: Unpack[_Attributes]
    ) -> None:
        _valid_type = (type(None), str, bytes, bool, int, float)
        if not isinstance(value, _valid_type):
            raise TypeError(f"Constants must be type: {_valid_type!r}")
        super().__init__(value, kind, **kwargs)


class Compare(Expr[bool], ast.Compare):
    """A comparison of two or more values.

    `left` is the first value in the comparison, `ops` the list of operators,
    and `comparators` the list of values after the first element in the
    comparison.
    """

    def __init__(
        self,
        left: Expr,
        ops: Sequence[ast.cmpop],
        comparators: Sequence[Expr],
        **kwargs: Unpack[_Attributes],
    ) -> None:
        super().__init__(
            Expr._cast(left), ops, [Expr._cast(c) for c in comparators], **kwargs
        )


class BinOp(Expr[T], ast.BinOp):
    """A binary operation (like addition or division).

    `op` is the operator, and `left` and `right` are any expression nodes.
    """

    def __init__(
        self,
        left: T | Expr[T],
        op: ast.operator,
        right: T | Expr[T],
        **kwargs: Unpack[_Attributes],
    ) -> None:
        super().__init__(Expr._cast(left), op, Expr._cast(right), **kwargs)


class BoolOp(Expr[T], ast.BoolOp):
    """A boolean operation, 'or' or 'and'.

    `op` is Or or And. `values` are the values involved. Consecutive operations
    with the same operator, such as a or b or c, are collapsed into one node
    with several values.

    This doesn't include `not`, which is a `UnaryOp`.
    """

    def __init__(
        self,
        op: ast.boolop,
        values: Sequence[ConstType | Expr],
        **kwargs: Unpack[_Attributes],
    ) -> None:
        super().__init__(op, [Expr._cast(v) for v in values], **kwargs)


class UnaryOp(Expr[T], ast.UnaryOp):
    """A unary operation.

    `op` is the operator, and `operand` any expression node.
    """

    def __init__(
        self, op: ast.unaryop, operand: Expr, **kwargs: Unpack[_Attributes]
    ) -> None:
        super().__init__(op, Expr._cast(operand), **kwargs)


class IfExp(Expr, ast.IfExp):
    """An expression such as `'a if b else c'`.

    `body` if `test` else `orelse`
    """

    def __init__(
        self, test: Expr, body: Expr, orelse: Expr, **kwargs: Unpack[_Attributes]
    ) -> None:
        super().__init__(
            Expr._cast(test), Expr._cast(body), Expr._cast(orelse), **kwargs
        )


class Tuple(Expr, ast.Tuple):
    """A tuple expression.

    `elts` is a list of expressions.
    """

    def __init__(
        self,
        elts: Sequence[Expr],
        ctx: ast.expr_context = LOAD,
        **kwargs: Unpack[_Attributes],
    ) -> None:
        super().__init__(elts=[Expr._cast(e) for e in elts], ctx=ctx, **kwargs)


class List(Expr, ast.List):
    """A tuple expression.

    `elts` is a list of expressions.
    """

    def __init__(
        self,
        elts: Sequence[Expr],
        ctx: ast.expr_context = LOAD,
        **kwargs: Unpack[_Attributes],
    ) -> None:
        super().__init__(elts=[Expr._cast(e) for e in elts], ctx=ctx, **kwargs)


class Set(Expr, ast.Set):
    """A tuple expression.

    `elts` is a list of expressions.
    """

    def __init__(self, elts: Sequence[Expr], **kwargs: Unpack[_Attributes]) -> None:
        super().__init__(elts=[Expr._cast(e) for e in elts], **kwargs)


class ExprTransformer(ast.NodeTransformer):
    """Transformer that converts an ast.expr into an `Expr`.

    Examples
    --------
    >>> tree = ast.parse("my_var > 11", mode="eval")
    >>> tree = ExprTransformer().visit(tree)  # transformed
    """

    _SUPPORTED_NODES = frozenset(
        k for k, v in globals().items() if isinstance(v, type) and issubclass(v, Expr)
    )

    # fmt: off
    @overload
    def visit(self, node: ast.expr) -> Expr: ...
    @overload
    def visit(self, node: PassedType) -> PassedType: ...
    # fmt: on
    def visit(self, node: ast.AST) -> ast.AST | None:
        """Visit a node in the tree, transforming into Expr."""
        if isinstance(
            node,
            (
                ast.cmpop,
                ast.operator,
                ast.boolop,
                ast.unaryop,
                ast.expr_context,
            ),
        ):
            # all operation types just get passed through
            return node

        # filter here for supported expression node types
        type_ = type(node).__name__

        if type_ not in ExprTransformer._SUPPORTED_NODES:
            raise SyntaxError(f"Type {type_!r} not supported")

        # providing fake lineno and col_offset here rather than using
        # ast.fill_missing_locations for typing purposes
        kwargs: dict[str, Any] = {"lineno": 1, "col_offset": 0}

        for name, field in ast.iter_fields(node):
            if isinstance(field, ast.expr):
                kwargs[name] = self.visit(field)
            elif isinstance(field, list):
                kwargs[name] = [self.visit(item) for item in field]
            else:
                kwargs[name] = field

        # return instance of Expr from this module corresponding to the node type
        return cast("Expr", globals()[type_](**kwargs))


class _ExprSerializer(ast.NodeVisitor):
    """Serializes an `Expr` into a string.

    Examples
    --------
    >>> expr = Expr.parse("a + b == c")
    >>> print(expr)
    'a + b == c'

    or ... using this visitor directly:

    >>> serializer = ExprSerializer()
    >>> serializer.visit(expr)
    >>> out = "".join(serializer.result)
    """

    def __init__(self, node: Expr | None = None) -> None:
        self._result: list[str] = []

        def write(*params: ast.AST | str) -> None:
            for item in params:
                if isinstance(item, ast.AST):
                    self.visit(item)
                elif item:
                    self._result.append(item)

        self.write = write

        if node is not None:
            self.visit(node)

    def __str__(self) -> str:
        return "".join(self._result)

    def visit_Name(self, node: ast.Name) -> None:
        self.write(node.id)

    def visit_Tuple(self, node: ast.Tuple) -> None:
        self.write(f"({', '.join(map(str, node.elts))})")

    def visit_Set(self, node: ast.Set) -> None:
        self.write("{" + ", ".join(map(str, node.elts)) + "}")

    def visit_List(self, node: ast.List) -> None:
        self.write(f"[{', '.join(map(str, node.elts))}]")

    def visit_ContextKey(self, node: ContextKey) -> None:
        return self.visit_Name(node)

    def visit_Constant(self, node: ast.Constant) -> None:
        self.write(repr(node.value))

    def visit_BoolOp(self, node: ast.BoolOp) -> Any:
        op = f" {_OPS[type(node.op)]} "
        for idx, value in enumerate(node.values):
            self.write((idx and op) or "", value)

    def visit_Compare(self, node: ast.Compare) -> None:
        self.visit(node.left)
        for op, right in zip(node.ops, node.comparators):
            self.write(f" {_OPS[type(op)]} ", right)

    def visit_BinOp(self, node: ast.BinOp) -> None:
        self.write(node.left, f" {_OPS[type(node.op)]} ", node.right)

    def visit_UnaryOp(self, node: ast.UnaryOp) -> None:
        sym = _OPS[type(node.op)]
        self.write(sym, " " if sym.isalpha() else "", node.operand)

    def visit_IfExp(self, node: ast.IfExp) -> Any:
        self.write(node.body, " if ", node.test, " else ", node.orelse)


OpType = Union[type[ast.operator], type[ast.cmpop], type[ast.boolop], type[ast.unaryop]]
_OPS: dict[OpType, str] = {
    # ast.boolop
    ast.Or: "or",
    ast.And: "and",
    # ast.cmpop
    ast.Eq: "==",
    ast.Gt: ">",
    ast.GtE: ">=",
    ast.In: "in",
    ast.Is: "is",
    ast.NotEq: "!=",
    ast.Lt: "<",
    ast.LtE: "<=",
    ast.NotIn: "not in",
    ast.IsNot: "is not",
    # ast.operator
    ast.BitOr: "|",
    ast.BitXor: "^",
    ast.BitAnd: "&",
    ast.LShift: "<<",
    ast.RShift: ">>",
    ast.Add: "+",
    ast.Sub: "-",
    ast.Mult: "*",
    ast.Div: "/",
    ast.Mod: "%",
    ast.FloorDiv: "//",
    ast.MatMult: "@",
    ast.Pow: "**",
    # ast.unaryop
    ast.Not: "not",
    ast.Invert: "~",
    ast.UAdd: "+",
    ast.USub: "-",
}


def _iter_names(expr: Expr) -> Iterator[str]:
    """Iterate all (nested) names used in the expression.

    Could be used to provide nicer error messages when eval() fails.
    """
    if isinstance(expr, Name):
        yield expr.id
    elif isinstance(expr, Expr):
        for _, val in ast.iter_fields(expr):
            val = val if isinstance(val, list) else [val]
            for v in val:
                yield from _iter_names(v)



import os
import sys
from enum import Enum, IntEnum


class KeyBindingSource(IntEnum):
    """Keybinding source enum."""

    APP = 0
    PLUGIN = 1
    USER = 2


class OperatingSystem(Enum):
    """Operating system enum."""

    UNKNOWN = 0
    WINDOWS = 1
    MACOS = 2
    LINUX = 3

    @staticmethod
    def current() -> "OperatingSystem":
        """Return the current operating system as enum."""
        return _CURRENT

    @property
    def is_windows(self) -> bool:
        """Returns True if the current operating system is Windows."""
        return _CURRENT == OperatingSystem.WINDOWS

    @property
    def is_linux(self) -> bool:
        """Returns True if the current operating system is Linux."""
        return _CURRENT == OperatingSystem.LINUX

    @property
    def is_mac(self) -> bool:
        """Returns True if the current operating system is MacOS."""
        return _CURRENT == OperatingSystem.MACOS


_CURRENT = OperatingSystem.UNKNOWN
if os.name == "nt":
    _CURRENT = OperatingSystem.WINDOWS
if sys.platform.startswith("linux"):
    _CURRENT = OperatingSystem.LINUX
elif sys.platform == "darwin":
    _CURRENT = OperatingSystem.MACOS



"""Generic application schema implemented in python."""

from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("app-model")
except PackageNotFoundError:  # pragma: no cover
    __version__ = "uninstalled"

from ._app import Application
from .registries._register import register_action
from .types import Action

__all__ = ["Action", "Application", "__version__", "register_action"]



from __future__ import annotations

from bisect import insort_left
from collections import defaultdict
from typing import TYPE_CHECKING, Callable, NamedTuple

from psygnal import Signal

from app_model.types import KeyBinding

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator, Mapping
    from typing import TypeVar

    from app_model import expressions
    from app_model.types import (
        Action,
        DisposeCallable,
        KeyBindingRule,
        KeyBindingSource,
    )

    CommandDecorator = Callable[[Callable], Callable]
    CommandCallable = TypeVar("CommandCallable", bound=Callable)


class _RegisteredKeyBinding(NamedTuple):
    """Internal object representing a fully registered keybinding."""

    keybinding: KeyBinding  # the keycode to bind to
    command_id: str  # the command to run
    weight: int  # the weight of the binding, for prioritization
    source: KeyBindingSource  # who defined the binding, for prioritization
    when: expressions.Expr | None = None  # condition to enable keybinding

    def __gt__(self, other: object) -> bool:
        if not isinstance(other, _RegisteredKeyBinding):
            return NotImplemented
        return (self.source, self.weight) > (other.source, other.weight)

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, _RegisteredKeyBinding):
            return NotImplemented
        return (self.source, self.weight) < (other.source, other.weight)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, _RegisteredKeyBinding):
            return NotImplemented
        return (self.source, self.weight) == (other.source, other.weight)


class KeyBindingsRegistry:
    """Registry for keybindings.

    Attributes
    ----------
    filter_keybinding : Callable[[KeyBinding], str] | None
        Optional function for applying additional `KeyBinding` filtering.
        Callable should accept a `KeyBinding` object and return an error message
        (`str`) if `KeyBinding` is rejected, or empty string otherwise.
    """

    registered = Signal()
    unregistered = Signal()

    def __init__(self) -> None:
        self._keymap = defaultdict[int, list[_RegisteredKeyBinding]](list)
        self._filter_keybinding: Callable[[KeyBinding], str] | None = None

    @property
    def _keybindings(self) -> Iterable[_RegisteredKeyBinding]:
        return (entry for entries in self._keymap.values() for entry in entries)

    @property
    def filter_keybinding(self) -> Callable[[KeyBinding], str] | None:
        """Return the `filter_keybinding`."""
        return self._filter_keybinding

    @filter_keybinding.setter
    def filter_keybinding(self, value: Callable[[KeyBinding], str] | None) -> None:
        if callable(value) or value is None:
            self._filter_keybinding = value
        else:
            raise TypeError("'filter_keybinding' must be a callable or None")

    def register_action_keybindings(self, action: Action) -> DisposeCallable | None:
        """Register all keybindings declared in `action.keybindings`.

        Parameters
        ----------
        action : Action
            The action to register keybindings for.

        Returns
        -------
        DisposeCallable | None
            A function that can be called to unregister the keybindings.  If no
            keybindings were registered, returns None.
        """
        if not (keybindings := action.keybindings):
            return None

        disposers: list[Callable[[], None]] = []
        msg: list[str] = []
        for keyb in keybindings:
            if action.enablement is not None:
                kwargs = keyb.model_dump()
                kwargs["when"] = (
                    action.enablement
                    if keyb.when is None
                    else action.enablement | keyb.when
                )
                _keyb = type(keyb)(**kwargs)
            else:
                _keyb = keyb

            try:
                if d := self.register_keybinding_rule(action.id, _keyb):
                    disposers.append(d)
            except ValueError as e:
                msg.append(str(e))
        if msg:
            raise ValueError(
                "The following keybindings were not valid:\n" + "\n".join(msg)
            )

        if not disposers:  # pragma: no cover
            return None

        def _dispose() -> None:
            for disposer in disposers:
                disposer()

        return _dispose

    def register_keybinding_rule(
        self, id: str, rule: KeyBindingRule
    ) -> DisposeCallable | None:
        """Register a new keybinding rule.

        Parameters
        ----------
        id : str
            Command identifier that should be run when the keybinding is triggered
        rule : KeyBindingRule
            KeyBinding information

        Returns
        -------
        Optional[DisposeCallable]
            A callable that can be used to unregister the keybinding
        """
        if plat_keybinding := rule._bind_to_current_platform():
            # list registry
            keybinding = KeyBinding.validate(plat_keybinding)
            if self._filter_keybinding:
                msg = self._filter_keybinding(keybinding)
                if msg:
                    raise ValueError(f"{keybinding}: {msg}")
            entry = _RegisteredKeyBinding(
                keybinding=keybinding,
                command_id=id,
                weight=rule.weight,
                when=rule.when,
                source=rule.source,
            )

            # inverse map registry
            entries = self._keymap[keybinding.to_int()]
            insort_left(entries, entry)

            self.registered.emit()

            def _dispose() -> None:
                # inverse map registry remove
                entries.remove(entry)
                self.unregistered.emit()
                if len(entries) == 0:
                    del self._keymap[keybinding.to_int()]

            return _dispose
        return None  # pragma: no cover

    def __iter__(self) -> Iterator[_RegisteredKeyBinding]:
        yield from self._keybindings

    def __len__(self) -> int:
        return sum(len(entries) for entries in self._keymap.values())

    def __repr__(self) -> str:
        name = self.__class__.__name__
        return f"<{name} at {hex(id(self))} ({len(self)} bindings)>"

    def get_keybinding(self, command_id: str) -> _RegisteredKeyBinding | None:
        """Return the first keybinding that matches the given command ID."""
        # TODO: improve me.
        matches = (kb for kb in self._keybindings if kb.command_id == command_id)
        sorted_matches = sorted(matches, key=lambda x: x.source, reverse=True)
        return next(iter(sorted_matches), None)

    def get_context_prioritized_keybinding(
        self, key: int, context: Mapping[str, object]
    ) -> _RegisteredKeyBinding | None:
        """
        Return the first keybinding that matches the given key sequence.

        The keybinding should be enabled given the context to be returned.

        Parameters
        ----------
        key : int
            The key sequence that represent the keybinding.
        context : Mapping[str, object]
            Variable context to parse the `when` expression to determine if the
            keybinding is enabled or not.

        Returns
        -------
        _RegisteredKeyBinding | None
            The keybinding found or None otherwise.

        """
        if key in self._keymap:
            for entry in reversed(self._keymap[key]):
                if entry.when is None or entry.when.eval(context):
                    return entry
        return None



from __future__ import annotations

from typing import TYPE_CHECKING

from qtpy.QtCore import Qt
from qtpy.QtWidgets import QMainWindow, QWidget

from app_model import Application

from ._qmenu import QModelMenuBar, QModelToolBar

if TYPE_CHECKING:
    from collections.abc import Collection, Mapping, Sequence


class QModelMainWindow(QMainWindow):
    """QMainWindow with app-model support."""

    def __init__(self, app: Application | str, parent: QWidget | None = None) -> None:
        super().__init__(parent)
        self._app = Application.get_or_create(app) if isinstance(app, str) else app

    def setModelMenuBar(
        self, menu_ids: Mapping[str, str] | Sequence[str | tuple[str, str]]
    ) -> QModelMenuBar:
        """Set the menu bar to a list of menu ids.

        Parameters
        ----------
        menu_ids : Mapping[str, str] | Sequence[str | tuple[str, str]]
            A mapping of menu ids to menu titles or a sequence of menu ids.
        """
        menu_bar = QModelMenuBar(menu_ids, self._app, self)
        self.setMenuBar(menu_bar)
        return menu_bar

    def addModelToolBar(
        self,
        menu_id: str,
        *,
        exclude: Collection[str] | None = None,
        area: Qt.ToolBarArea | None = None,
        toolbutton_style: Qt.ToolButtonStyle = Qt.ToolButtonStyle.ToolButtonIconOnly,
    ) -> QModelToolBar:
        """Add a tool bar to the main window."""
        toolbar = QModelToolBar(menu_id, self._app, exclude=exclude, parent=self)
        toolbar.setToolButtonStyle(toolbutton_style)
        if area is not None:
            self.addToolBar(area, toolbar)
        else:
            self.addToolBar(toolbar)
        return toolbar



from __future__ import annotations

import os
import sys
from typing import TYPE_CHECKING

import pytest

from app_model import Application
from app_model.expressions import Context
from app_model.types import Action
from app_model.types._menu_rule import MenuRule

if TYPE_CHECKING:
    from conftest import FullApp


def test_app_create() -> None:
    assert Application.get_app("my_app") is None
    app = Application("my_app")
    assert Application.get_app("my_app") is app

    # NOTE: for some strange reason, this test fails if I move this line
    # below the error assertion below... I don't know why.
    assert Application.get_or_create("my_app") is app

    with pytest.raises(ValueError, match="Application 'my_app' already exists"):
        Application("my_app")

    assert repr(app) == "Application('my_app')"
    Application.destroy("my_app")


def test_app(full_app: FullApp) -> None:
    app = full_app

    app.commands.execute_command(app.Commands.OPEN)
    app.mocks.open.assert_called_once()
    app.commands.execute_command(app.Commands.COPY)
    app.mocks.copy.assert_called_once()
    app.commands.execute_command(app.Commands.PASTE)
    app.mocks.paste.assert_called_once()


def test_sorting(full_app: FullApp) -> None:
    groups = list(full_app.menus.iter_menu_groups(full_app.Menus.EDIT))
    assert len(groups) == 3
    [_g0, g1, g2] = groups
    assert all(i.group == "1_undo_redo" for i in g1)
    assert all(i.group == "2_copy_paste" for i in g2)

    assert [i.command.title for i in g1] == ["Undo", "Redo"]
    assert [i.command.title for i in g2] == ["Copy", "Paste"]


def test_action_import_by_string(full_app: FullApp) -> None:
    """the REDO command is declared as a string in the conftest.py file

    This tests that it can be lazily imported at callback runtime and executed
    """
    assert "fake_module" not in sys.modules
    assert full_app.commands.execute_command(full_app.Commands.REDO).result()
    assert "fake_module" in sys.modules
    full_app.mocks.redo.assert_called_once()

    # tests what happens when the module cannot be found
    with pytest.raises(
        ModuleNotFoundError, match="Command 'unimportable' was not importable"
    ):
        full_app.commands.execute_command(full_app.Commands.UNIMPORTABLE)
    # the second time we try within a session, nothing should happen
    full_app.commands.execute_command(full_app.Commands.UNIMPORTABLE)

    # tests what happens when the object is not callable cannot be found
    with pytest.raises(
        TypeError,
        match=r"Command 'not\.callable' did not resolve to a callble object",
    ):
        full_app.commands.execute_command(full_app.Commands.NOT_CALLABLE)
    # the second time we try within a session, nothing should happen
    full_app.commands.execute_command(full_app.Commands.NOT_CALLABLE)


def test_action_raises_exception(full_app: FullApp) -> None:
    result = full_app.commands.execute_command(full_app.Commands.RAISES)
    with pytest.raises(ValueError):
        result.result()

    # the function that raised the exception is `_raise_an_error` in conftest.py
    assert str(result.exception()) == "This is an error"

    assert not full_app.raise_synchronous_exceptions
    full_app.raise_synchronous_exceptions = True
    assert full_app.raise_synchronous_exceptions

    with pytest.raises(ValueError):
        full_app.commands.execute_command(full_app.Commands.RAISES)


def test_app_context() -> None:
    app = Application("app1")
    assert isinstance(app.context, Context)
    Application.destroy("app1")
    assert app.context["is_windows"] == (os.name == "nt")
    assert "is_mac" in app.context
    assert "is_linux" in app.context

    app = Application("app2", context={"a": 1})
    assert isinstance(app.context, Context)
    assert app.context["a"] == 1
    Application.destroy("app2")

    app = Application("app3", context=Context({"a": 1}))
    assert isinstance(app.context, Context)
    assert app.context["a"] == 1
    Application.destroy("app3")

    with pytest.raises(TypeError, match="context must be a Context or MutableMapping"):
        Application("app4", context=1)  # type: ignore[arg-type]


def test_register_actions() -> None:
    app = Application("app5")
    actions = app.registered_actions
    assert not actions
    dispose = app.register_action(
        "my_action", title="My Action", callback=lambda: None, menus=["Window"]
    )
    assert "my_action" in actions
    assert isinstance(action := actions["my_action"], Action)
    assert action.menus == [MenuRule(id="Window")]
    dispose()
    assert "my_action" not in actions
    assert not actions



from __future__ import annotations

from typing import TYPE_CHECKING, overload

from app_model.types import Action

if TYPE_CHECKING:
    from typing import Any, Callable, Literal, TypeVar

    from app_model import Application, expressions
    from app_model.types import (
        DisposeCallable,
        Icon,  # noqa: F401 ... used in type hints for docs
        IconOrDict,
        KeyBindingRuleOrDict,
        MenuRuleOrDict,
    )

    CommandCallable = TypeVar("CommandCallable", bound=Callable[..., Any])
    CommandDecorator = Callable[[Callable], Callable]


@overload
def register_action(
    app: Application | str, id_or_action: Action
) -> DisposeCallable: ...


@overload
def register_action(
    app: Application | str,
    id_or_action: str,
    title: str,
    *,
    callback: Literal[None] = ...,
    category: str | None = ...,
    tooltip: str | None = ...,
    icon: IconOrDict | None = ...,
    enablement: expressions.Expr | None = ...,
    menus: list[MenuRuleOrDict] | None = ...,
    keybindings: list[KeyBindingRuleOrDict] | None = ...,
    palette: bool = True,
) -> CommandDecorator: ...


@overload
def register_action(
    app: Application | str,
    id_or_action: str,
    title: str,
    *,
    callback: Callable[..., Any],
    category: str | None = ...,
    tooltip: str | None = ...,
    icon: IconOrDict | None = ...,
    enablement: expressions.Expr | None = ...,
    menus: list[MenuRuleOrDict] | None = ...,
    keybindings: list[KeyBindingRuleOrDict] | None = ...,
    palette: bool = True,
) -> DisposeCallable: ...


def register_action(
    app: Application | str,
    id_or_action: str | Action,
    title: str | None = None,
    *,
    callback: Callable[..., Any] | None = None,
    category: str | None = None,
    tooltip: str | None = None,
    icon: IconOrDict | None = None,
    enablement: expressions.Expr | None = None,
    menus: list[MenuRuleOrDict] | None = None,
    keybindings: list[KeyBindingRuleOrDict] | None = None,
    palette: bool = True,
) -> CommandDecorator | DisposeCallable:
    """Register an action.

    This is a functional form of the
    [`Application.register_action()`][app_model.Application.register_action] method.
    It accepts various overloads to allow for a more concise syntax.  See examples
    below.

    An `Action` is the "complete" representation of a command.  The command is the
    function/callback itself, and an action also includes information about where and
    whether it appears in menus and optional keybinding rules.  Since, most of the
    arguments to this function are simply passed through to the `Action` constructor,
    see also docstrings for:

    - [`Action`][app_model.types.Action]
    - [`CommandRule`][app_model.types.CommandRule]
    - [`MenuRule`][app_model.types.MenuRule]
    - [`KeyBindingRule`][app_model.types.KeyBindingRule]

    Parameters
    ----------
    app: Application | str
        The app in which to register the action. If a string, the app is retrieved
        or created as necessary using
        [`Application.get_or_create(app)`][app_model.Application.get_or_create].
    id_or_action : str | Action
        Either a complete Action object or a string id of the command being registered.
        If an `Action` object is provided, then all other arguments are ignored.
    title : str | None
        Title by which the command is represented in the UI. Required when
        `id_or_action` is a string.
    callback : CommandHandler | None
        Callable object that executes this command, by default None. If not provided,
        a decorator is returned that can be used to decorate a function that executes
        this action.
    category : str | None
        Category string by which the command may be grouped in the UI, by default None
    tooltip : str | None
        Tooltip to show when hovered., by default None
    icon : Icon | None
        [`Icon`][app_model.types.Icon] used to represent this command,
        e.g. on buttons or in menus. by default None
    enablement : expressions.Expr | None
        Condition which must be true to enable the command in in the UI,
        by default None
    menus : list[MenuRuleOrDict] | None
        List of [`MenuRule`][app_model.types.MenuRule] or kwarg `dicts` containing menu
        placements for this action, by default None
    keybindings : list[KeyBindingRuleOrDict] | None
        List of [`KeyBindingRule`][app_model.types.KeyBindingRule] or kwargs `dicts`
        containing default keybindings for this action, by default None
    palette : bool
        Whether to adds this command to the Command Palette, by default True

    Returns
    -------
    CommandDecorator
        If `callback` is not provided, then a decorator is returned that can be used to
        decorate a function as the executor of the command.
    DisposeCallable
        If `callback` is provided, or `id_or_action` is an `Action` object, then a
        function is returned that may be used to unregister the action.

    Raises
    ------
    ValueError
        If `id_or_action` is a string and `title` is not provided.
    TypeError
        If `id_or_action` is not a string or an `Action` object.

    Examples
    --------
    This function can be used directly or as a decorator, and accepts arguments in
    various forms.

    ## Passing an existing Action object

    When the `id_or_action` argument is an instance of `app_model.Action`, then
    all other arguments are ignored, the action object is registered directly, and the
    return value is a function that may be used to unregister the action is returned.

    ```python
    from app_model import Application, Action, register_action

    app = Application.get_or_create("myapp")
    action = Action("my_action", title="My Action", callback=lambda: print("hi"))
    register_action(app, action)

    app.commands.execute_command("my_action")  # prints "hi"
    ```

    ## Creating a new Action

    When the `id_or_action` argument is a string, it is interpreted as the `id`
    of the command being registered, in which case `title` must then also be provided.
    All other arguments are optional, but may be used to customize the action being
    created (with keybindings, menus, icons, etc).

    ```python
    register_action(
        app,
        "my_action2",
        title="My Action2",
        callback=lambda: print("hello again!"),
    )

    app.commands.execute_command("my_action2")  # prints "hello again!"
    ```

    ## Usage as a decorator

    If `callback` is not provided, then a decorator is returned that can be used
    decorate a function as the executor of the command:

    ```python
    @register_action(app, "my_action3", title="My Action3")
    def my_action3():
        print("hello again, again!")


    app.commands.execute_command("my_action3")  # prints "hello again, again!"
    ```

    ## Passing app as a string

    Note that in all of the above examples, the first `app` argument may be either an
    instance of an [`Application`][app_model.Application] object, or a string name of
    an application.  If a string is provided, then the application is retrieved or
    created as necessary using
    [`Application.get_or_create()`][app_model.Application.get_or_create].

    ```python
    register_action(
        "myapp",  # app name instead of Application instance
        "my_action4",
        title="My Action4",
        callback=lambda: print("hello again, again, again!"),
    )
    ```
    """
    if isinstance(id_or_action, Action):
        return _register_action_obj(app, id_or_action)
    if isinstance(id_or_action, str):
        if not title:
            raise ValueError("'title' is required when 'id' is a string")
        return _register_action_str(
            app=app,
            id=id_or_action,
            title=title,
            category=category,
            tooltip=tooltip,
            icon=icon,
            enablement=enablement,
            callback=callback,
            palette=palette,
            menus=menus,
            keybindings=keybindings,
        )
    raise TypeError(
        f"'id_or_action' must be a string or an Action not {type(id_or_action)}"
    )


def _register_action_str(
    app: Application | str, **kwargs: Any
) -> CommandDecorator | DisposeCallable:
    """Create and register an Action with a string id and title.

    Helper for `register_action()`.

    If `kwargs['callback']` is a callable, a complete `Action` is created
    (thereby performing type validation and casting) and registered with the
    corresponding registries. Otherwise a decorator returned that can be used
    to decorate the callable that executes the action.
    """
    if kwargs.get("callback") is not None:
        return _register_action_obj(app, Action(**kwargs))

    def decorator(command: CommandCallable, **k: Any) -> CommandCallable:
        if not callable(command):
            raise TypeError(
                "@register_action decorator must be passed a callable object"
            )
        _register_action_obj(app, Action(**{**kwargs, **k, "callback": command}))
        return command

    decorator.__doc__ = f"Decorate function as callback for command {kwargs['id']!r}"
    return decorator


def _register_action_obj(app: Application | str, action: Action) -> DisposeCallable:
    """Register an Action object. Return a function that unregisters the action.

    Helper for `register_action()`.
    """
    from app_model._app import Application

    app = app if isinstance(app, Application) else Application.get_or_create(app)
    return app._register_action_obj(action)



from typing import TYPE_CHECKING, Callable, Generic, Optional, TypeVar, Union

from pydantic import Field, field_validator

from ._command_rule import CommandRule
from ._keybinding_rule import KeyBindingRule
from ._menu_rule import MenuRule
from ._utils import _validate_python_name

# maintain runtime compatibility with older typing_extensions
if TYPE_CHECKING:
    from typing_extensions import ParamSpec

    P = ParamSpec("P")
else:
    try:
        from typing_extensions import ParamSpec

        P = ParamSpec("P")
    except ImportError:
        P = TypeVar("P")
R = TypeVar("R")


class Action(CommandRule, Generic[P, R]):
    """An Action is a callable object with menu placement, keybindings, and metadata.

    This is the "complete" representation of a command.  Including a pointer to the
    actual callable object, as well as any additional menu and keybinding rules.
    Most commands and menu items will be represented by Actions, and registered using
    `register_action`.
    """

    callback: Union[Callable[P, R], str] = Field(
        ...,
        description="A function to call when the associated command id is executed. "
        "If a string is provided, it must be a fully qualified name to a callable "
        "python object. This usually takes the form of "
        "`{obj.__module__}:{obj.__qualname__}` "
        "(e.g. `my_package.a_module:some_function`)",
    )
    menus: Optional[list[MenuRule]] = Field(
        default=None,
        description="(Optional) Menus to which this action should be added.  Note that "
        "menu items in the sequence may be supplied as a plain string, which will "
        "be converted to a `MenuRule` with the string as the `id` field.",
    )
    keybindings: Optional[list[KeyBindingRule]] = Field(
        default=None,
        description="(Optional) Default keybinding(s) that will trigger this command.",
    )
    palette: bool = Field(
        default=True,
        description="Whether to add this command to the global Command Palette "
        "during registration.",
    )

    @field_validator("callback")
    def _validate_callback(callback: object) -> Union[Callable, str]:
        """Assert that `callback` is a callable or valid fully qualified name."""
        if callable(callback):
            return callback
        elif isinstance(callback, str):
            return _validate_python_name(str(callback))
        raise TypeError("callback must be a callable or a string")  # pragma: no cover



from app_model import Application
from app_model.backends.qt import QModelMainWindow

from .actions import ACTIONS
from .constants import MenuId


class MyApp(Application):
    def __init__(self) -> None:
        super().__init__("my_application")

        # ACTIONS is a list of Action objects.
        for action in ACTIONS:
            self.register_action(action)

        self._main_window = QModelMainWindow(app=self)
        # This will build a menu bar based on these menus
        self._main_window.setModelMenuBar([MenuId.FILE, MenuId.EDIT])

    def show(self) -> None:
        """Show the app"""
        self._main_window.show()



from __future__ import annotations

from concurrent.futures import Future, ThreadPoolExecutor
from typing import TYPE_CHECKING, Any, Callable, Generic, TypeVar, cast

from in_n_out import Store
from psygnal import Signal

# maintain runtime compatibility with older typing_extensions
if TYPE_CHECKING:
    from collections.abc import Iterator

    from typing_extensions import ParamSpec

    from app_model.types import Action, DisposeCallable

    P = ParamSpec("P")
else:
    try:
        from typing_extensions import ParamSpec

        P = ParamSpec("P")
    except ImportError:
        P = TypeVar("P")


R = TypeVar("R")


class RegisteredCommand(Generic[P, R]):
    """Small object to represent a command in the CommandsRegistry.

    Used internally by the CommandsRegistry.

    This helper class allows us to cache the dependency-injected variant of the
    command, so that type resolution and dependency injection is performed only
    once.
    """

    __slots__ = (
        "_initialized",
        "_injected_callback",
        "_injection_store",
        "_resolved_callback",
        "callback",
        "id",
        "title",
    )

    def __init__(
        self,
        id: str,
        callback: Callable[P, R] | str,
        title: str,
        store: Store | None = None,
    ) -> None:
        self.id = id
        self.callback = callback
        self.title = title
        self._injection_store: Store = store or Store.get_store()
        self._resolved_callback = callback if callable(callback) else None
        self._injected_callback: Callable[P, R] | None = None
        self._initialized = True

    def __setattr__(self, name: str, value: Any) -> None:
        """Object is immutable after initialization."""
        if getattr(self, "_initialized", False):
            raise AttributeError("RegisteredCommand object is immutable.")
        super().__setattr__(name, value)

    @property
    def resolved_callback(self) -> Callable[P, R]:
        """Return the resolved command callback.

        This property is cached, so the callback types are only resolved once.
        """
        if self._resolved_callback is None:
            from app_model.types._utils import import_python_name

            try:
                cb = import_python_name(str(self.callback))
            except ImportError as e:
                object.__setattr__(self, "_resolved_callback", lambda *a, **k: None)
                raise type(e)(
                    f"Command pointer {self.callback!r} registered for Command "
                    f"{self.id!r} was not importable: {e}"
                ) from e

            if not callable(cb):
                # don't try to import again, just create a no-op
                object.__setattr__(self, "_resolved_callback", lambda *a, **k: None)
                raise TypeError(
                    f"Command pointer {self.callback!r} registered for Command "
                    f"{self.id!r} did not resolve to a callble object."
                )
            object.__setattr__(self, "_resolved_callback", cb)
        return cast("Callable[P, R]", self._resolved_callback)

    @property
    def run_injected(self) -> Callable[P, R]:
        """Return the command callback with dependencies injected.

        This property is cached, so the injected version is only created once.
        """
        if self._injected_callback is None:
            cb = self._injection_store.inject(self.resolved_callback, processors=True)
            object.__setattr__(self, "_injected_callback", cb)
        return cast("Callable[P, R]", self._injected_callback)


class CommandsRegistry:
    """Registry for commands (callable objects)."""

    registered = Signal(str)

    def __init__(
        self,
        injection_store: Store | None = None,
        raise_synchronous_exceptions: bool = False,
    ) -> None:
        self._commands: dict[str, RegisteredCommand] = {}
        self._injection_store = injection_store
        self._raise_synchronous_exceptions = raise_synchronous_exceptions

    def register_action(self, action: Action) -> DisposeCallable:
        """Register an Action object.

        This is a convenience method that registers the action's callback
        with the action's ID and title using `register_command`.

        Parameters
        ----------
        action: Action
            Action to register

        Returns
        -------
        DisposeCallable
            A function that can be called to unregister the action.
        """
        return self.register_command(action.id, action.callback, action.title)

    def register_command(
        self, id: str, callback: Callable[P, R] | str, title: str
    ) -> DisposeCallable:
        """Register a callable as the handler for command `id`.

        Parameters
        ----------
        id : CommandId
            Command identifier
        callback : Callable
            Callable to be called when the command is executed
        title : str
            Title for the command.

        Returns
        -------
        DisposeCallable
            A function that can be called to unregister the command.
        """
        if id in self._commands:
            raise ValueError(
                f"Command {id!r} already registered with callback "
                f"{self._commands[id].callback!r} (new callback: {callback!r})"
            )

        cmd = RegisteredCommand(id, callback, title, self._injection_store)
        self._commands[id] = cmd

        def _dispose() -> None:
            self._commands.pop(id, None)

        self.registered.emit(id)
        return _dispose

    def __iter__(self) -> Iterator[tuple[str, RegisteredCommand]]:
        yield from self._commands.items()

    def __len__(self) -> int:
        return len(self._commands)

    def __contains__(self, id: str) -> bool:
        return id in self._commands

    def __repr__(self) -> str:
        name = self.__class__.__name__
        return f"<{name} at {hex(id(self))} ({len(self._commands)} commands)>"

    def __getitem__(self, id: str) -> RegisteredCommand:
        """Retrieve commands registered under a given ID."""
        if id not in self._commands:
            raise KeyError(f"Command {id!r} not registered")
        return self._commands[id]

    def execute_command(
        self,
        id: str,
        *args: Any,
        execute_asynchronously: bool = False,
        **kwargs: Any,
    ) -> Future:
        """Execute a registered command.

        Parameters
        ----------
        id : CommandId
            ID of the command to execute
        *args: Any
            Positional arguments to pass to the command
        execute_asynchronously : bool
            Whether to execute the command asynchronously in a thread,
            by default `False`.  Note that *regardless* of this setting,
            the return value will implement the `Future` API (so it's necessary)
            to call `result()` on the returned object.  Eventually, this will
            default to True, but we need to solve `ensure_main_thread` Qt threading
            issues first
        **kwargs: Any
            Keyword arguments to pass to the command

        Returns
        -------
        Future: concurrent.futures.Future
            Future object containing the result of the command

        Raises
        ------
        KeyError
            If the command is not registered or has no callbacks.
        """
        try:
            cmd = self[id].run_injected
        except KeyError as e:
            raise KeyError(f"Command {id!r} not registered") from e  # pragma: no cover

        if execute_asynchronously:
            with ThreadPoolExecutor() as executor:
                return executor.submit(cmd, *args, **kwargs)

        future: Future = Future()
        try:
            future.set_result(cmd(*args, **kwargs))
        except Exception as e:
            if self._raise_synchronous_exceptions:
                # note, the caller of this function can also achieve this by
                # calling `future.result()` on the returned future object.
                raise e
            future.set_exception(e)

        return future

    def __str__(self) -> str:
        lines = [f"{id_!r:<32} -> {cmd.title!r}" for id_, cmd in self]
        return "\n".join(lines)



import itertools
import sys

import pytest
from pydantic import BaseModel

from app_model.types import (
    KeyBinding,
    KeyBindingRule,
    KeyCode,
    KeyMod,
    SimpleKeyBinding,
)
from app_model.types._constants import OperatingSystem
from app_model.types._keys import KeyChord, KeyCombo, StandardKeyBinding

MAC = sys.platform == "darwin"


@pytest.mark.parametrize("use_symbols", [True, False])
@pytest.mark.parametrize(
    ("os", "joinchar", "expected_use_symbols", "expected_non_use_symbols"),
    [
        (OperatingSystem.WINDOWS, "+", "⊞+A", "Win+A"),
        (OperatingSystem.LINUX, "-", "Super-A", "Super-A"),
        (OperatingSystem.MACOS, "", "⌘A", "CmdA"),
    ],
)
def test_simple_keybinding_to_text(
    use_symbols: bool,
    os: OperatingSystem,
    joinchar: str,
    expected_use_symbols: str,
    expected_non_use_symbols: str,
) -> None:
    kb = SimpleKeyBinding.from_str("Meta+A")
    expected = expected_non_use_symbols
    if use_symbols:
        expected = expected_use_symbols
    assert kb.to_text(os=os, use_symbols=use_symbols, joinchar=joinchar) == expected


@pytest.mark.parametrize("use_symbols", [True, False])
@pytest.mark.parametrize(
    ("os", "joinchar", "expected_use_symbols", "expected_non_use_symbols"),
    [
        (
            OperatingSystem.WINDOWS,
            "+",
            "Ctrl+A ⇧+[ Alt+/ ⊞+9",
            "Ctrl+A Shift+[ Alt+/ Win+9",
        ),
        (
            OperatingSystem.LINUX,
            "-",
            "Ctrl-A ⇧-[ Alt-/ Super-9",
            "Ctrl-A Shift-[ Alt-/ Super-9",
        ),
        (OperatingSystem.MACOS, "", "⌃A ⇧[ ⌥/ ⌘9", "ControlA Shift[ Option/ Cmd9"),
    ],
)
def test_keybinding_to_text(
    use_symbols: bool,
    os: OperatingSystem,
    joinchar: str,
    expected_use_symbols: str,
    expected_non_use_symbols: str,
) -> None:
    kb = KeyBinding.from_str("Ctrl+A Shift+[ Alt+/ Meta+9")
    expected = expected_non_use_symbols
    if use_symbols:
        expected = expected_use_symbols
    assert kb.to_text(os=os, use_symbols=use_symbols, joinchar=joinchar) == expected


@pytest.mark.parametrize("key", list("ADgf`]/,"))
@pytest.mark.parametrize("mod", ["ctrl", "shift", "alt", "meta", None])
def test_simple_keybinding_single_mod(mod: str, key: str) -> None:
    _mod = f"{mod}+" if mod else ""
    kb = SimpleKeyBinding.from_str(f"{_mod}{key}")
    assert str(kb).lower() == f"{_mod}{key}".lower()
    assert not kb.is_modifier_key()

    # we can compare it with another SimpleKeyBinding
    # using validate method just for test coverage... will pass to from_str
    assert kb == SimpleKeyBinding._parse_input(f"{_mod}{key}")
    # or with a string
    assert kb == f"{_mod}{key}"
    assert kb != ["A", "B"]  # check type error during comparison

    # round trip to int
    assert isinstance(kb.to_int(), KeyCombo)
    # using validate method just for test coverage... will pass to from_int
    assert SimpleKeyBinding._parse_input(int(kb)) == kb
    assert SimpleKeyBinding._parse_input(kb) == kb

    # first part of a Keybinding is a simple keybinding
    as_full_kb = KeyBinding.validate(kb)
    assert as_full_kb.part0 == kb
    assert KeyBinding.validate(int(kb)).part0 == kb

    assert int(as_full_kb) == int(kb)


def test_simple_keybinding_multi_mod() -> None:
    # here we're also testing that cmd and win get cast to 'KeyMod.CtrlCmd'

    kb = SimpleKeyBinding.from_str("cmd+shift+A")
    assert not kb.is_modifier_key()
    assert int(kb) & KeyMod.CtrlCmd | KeyMod.Shift

    kb = SimpleKeyBinding.from_str("win+shift+A")
    assert not kb.is_modifier_key()
    assert int(kb) & KeyMod.CtrlCmd | KeyMod.Shift

    kb = SimpleKeyBinding.from_str("win")  # just a modifier
    assert kb.is_modifier_key()


controls = ["ctrl", "control", "ctl", "⌃", "^"]
shifts = ["shift", "⇧"]
alts = ["alt", "opt", "option", "⌥"]
metas = ["meta", "super", "cmd", "command", "⌘", "win", "windows", "⊞"]
delimiters = ["+", "-"]
key = ["A"]
combos = [
    delim.join(x)
    for delim, *x in itertools.product(delimiters, controls, shifts, alts, metas, key)
]


@pytest.mark.parametrize("key", combos)
def test_keybinding_parser(key: str) -> None:
    # Test all the different ways to write the modifiers
    assert str(KeyBinding.from_str(key)) == "Ctrl+Shift+Alt+Meta+A"


def test_chord_keybinding() -> None:
    kb = KeyBinding.from_str("Shift+A Cmd+9")
    assert len(kb) == 2
    assert kb != "Shift+A Cmd+9"  # comparison with string considered anti-pattern
    assert kb == KeyBinding.from_str("Shift+A Cmd+9")
    assert kb.part0 == SimpleKeyBinding(shift=True, key=KeyCode.KeyA)
    assert kb.part0 == "Shift+A"
    assert str(kb) in repr(kb)
    # round trip to int
    assert isinstance(kb.to_int(), KeyChord)
    # using validate method just for test coverage... will pass to from_int
    assert KeyBinding.validate(int(kb)) == kb
    assert KeyBinding.validate(kb) == kb


def test_in_dict() -> None:
    a = SimpleKeyBinding.from_str("Shift+A")
    b = KeyBinding.from_str("Shift+B")

    try:
        kbs = {
            a: 0,
            b: 1,
        }
    except TypeError as e:
        if str(e).startswith("unhashable type"):
            pytest.fail(f"keybinds not hashable: {e}")
        else:
            raise e

    assert kbs[a] == 0
    assert kbs[b] == 1

    new_a = KeyBinding.from_str("Shift+A")

    with pytest.raises(KeyError):
        kbs[new_a]


def test_in_model() -> None:
    class M(BaseModel):
        key: KeyBinding

    m = M(key="Shift+A B")
    assert m.model_dump_json() == '{"key":"Shift+A B"}'


def test_standard_keybindings() -> None:
    class M(BaseModel):
        key: KeyBindingRule

    m = M(key=StandardKeyBinding.Copy)
    assert m.key.primary == KeyMod.CtrlCmd | KeyCode.KeyC



from typing import TYPE_CHECKING
from unittest.mock import Mock

import pytest

from app_model.backends.qt import QCommandRuleAction, QMenuItemAction
from app_model.types import (
    Action,
    CommandRule,
    KeyBindingRule,
    KeyBindingSource,
    KeyCode,
    MenuItem,
    ToggleRule,
)

if TYPE_CHECKING:
    from app_model import Application
    from conftest import FullApp


@pytest.mark.usefixtures("qapp")
def test_cache_qaction(full_app: "FullApp") -> None:
    action = next(
        i for k, items in full_app.menus for i in items if isinstance(i, MenuItem)
    )
    a1 = QMenuItemAction.create(action, full_app)
    a2 = QMenuItemAction.create(action, full_app)
    assert a1 is a2
    assert repr(a1).startswith("QMenuItemAction")


@pytest.mark.usefixtures("qapp")
def test_toggle_qaction(simple_app: "Application") -> None:
    mock = Mock()
    x = False

    def current() -> bool:
        mock()
        return x

    def _toggle() -> None:
        nonlocal x
        x = not x

    action = Action(
        id="test.toggle",
        title="Test toggle",
        toggled=ToggleRule(get_current=current),
        callback=_toggle,
    )
    simple_app.register_action(action)

    a1 = QCommandRuleAction(action, simple_app)
    mock.assert_called_once()
    mock.reset_mock()

    assert a1.isCheckable()
    assert not a1.isChecked()

    a1.trigger()
    assert a1.isChecked()
    assert x

    a1.trigger()
    assert not a1.isChecked()
    assert not x

    x = True
    a1._refresh()
    mock.assert_called_once()
    assert a1.isChecked()


def test_icon_visible_in_menu(qapp, simple_app: "Application") -> None:
    rule = CommandRule(id="test", title="Test", icon_visible_in_menu=False)
    q_action = QCommandRuleAction(command_rule=rule, app=simple_app)
    assert not q_action.isIconVisibleInMenu()


@pytest.mark.usefixtures("qapp")
@pytest.mark.parametrize(
    ("tooltip", "expected_tooltip"),
    [
        ("", "Test tooltip"),
        ("Test action with a tooltip", "Test action with a tooltip"),
    ],
)
def test_tooltip(
    simple_app: "Application", tooltip: str, expected_tooltip: str
) -> None:
    action = Action(
        id="test.tooltip", title="Test tooltip", tooltip=tooltip, callback=lambda: None
    )
    simple_app.register_action(action)
    q_action = QCommandRuleAction(action, simple_app)
    assert q_action.toolTip() == expected_tooltip


@pytest.mark.usefixtures("qapp")
@pytest.mark.parametrize(
    ("tooltip", "tooltip_with_keybinding", "tooltip_without_keybinding"),
    [
        ("", "Test keybinding tooltip (K)", "Test keybinding tooltip"),
        (
            "Test action with a tooltip and a keybinding",
            "Test action with a tooltip and a keybinding (K)",
            "Test action with a tooltip and a keybinding",
        ),
    ],
)
def test_keybinding_in_tooltip(
    simple_app: "Application",
    tooltip: str,
    tooltip_with_keybinding: str,
    tooltip_without_keybinding: str,
) -> None:
    action = Action(
        id="test.keybinding.tooltip",
        title="Test keybinding tooltip",
        callback=lambda: None,
        tooltip=tooltip,
        keybindings=[KeyBindingRule(primary=KeyCode.KeyK)],
    )
    simple_app.register_action(action)

    # check initial action instance shows keybinding info in its tooltip if available
    q_action = QCommandRuleAction(action, simple_app)
    assert q_action.toolTip() == tooltip_with_keybinding

    # check setting tooltip manually removes keybinding info
    q_action.setToolTip(tooltip)
    assert q_action.toolTip() == tooltip_without_keybinding


@pytest.mark.usefixtures("qapp")
def test_update_keybinding_in_tooltip(
    simple_app: "Application",
) -> None:
    action = Action(
        id="test.update.keybinding.tooltip",
        title="Test update keybinding tooltip",
        callback=lambda: None,
        tooltip="Initial tooltip",
        keybindings=[KeyBindingRule(primary=KeyCode.KeyK)],
    )
    dispose1 = simple_app.register_action(action)

    q_action = QCommandRuleAction(action, simple_app)
    assert q_action.toolTip() == "Initial tooltip (K)"

    # Update the keybinding
    dispose2 = simple_app.keybindings.register_keybinding_rule(
        "test.update.keybinding.tooltip",
        KeyBindingRule(primary=KeyCode.KeyL, source=KeyBindingSource.USER),
    )
    q_action._update_keybinding()
    assert q_action.toolTip() == "Initial tooltip (L)"

    dispose2()
    q_action._update_keybinding()
    assert q_action.toolTip() == "Initial tooltip (K)"

    dispose1()
    q_action._update_keybinding()
    assert q_action.toolTip() == "Initial tooltip"



from typing import TYPE_CHECKING, Optional

from qtpy.QtWidgets import QKeySequenceEdit

from ._qkeymap import qkeysequence2modelkeybinding

if TYPE_CHECKING:
    from app_model.types import KeyBinding


class QModelKeyBindingEdit(QKeySequenceEdit):
    """Editor for a KeyBinding instance.

    This is a QKeySequenceEdit with a method that converts the current
    keySequence to an app_model KeyBinding instance.
    """

    def keyBinding(self) -> Optional["KeyBinding"]:
        """Return app_model KeyBinding instance for the current keySequence."""
        if self.keySequence().isEmpty():
            return None
        return qkeysequence2modelkeybinding(self.keySequence())



from typing import Any


def on_page_markdown(md: str, **kwargs: Any) -> str:
    T = "::: app_model.types"
    T2 = T + "\n\toptions:\n\t\tdocstring_section_style: table"
    return md.replace(T, T2)



from typing import Any, Optional, TypedDict, Union

from pydantic import Field, model_validator

from ._base import _BaseModel


class Icon(_BaseModel):
    """Icons used to represent commands, or submenus.

    May provide both a light and dark variant.  If only one is provided, it is used
    in all theme types.
    """

    dark: Optional[str] = Field(
        default=None,
        description="Icon path when a dark theme is used. These may be "
        "[iconify keys](https://icon-sets.iconify.design), such as "
        "`fa6-solid:arrow-down`, or "
        "[superqt.fonticon](https://pyapp-kit.github.io/superqt/utilities/fonticon/)"
        " keys, such as `fa6s.arrow_down`",
    )
    light: Optional[str] = Field(
        default=None,
        description="Icon path when a light theme is used. These may be "
        "[iconify keys](https://icon-sets.iconify.design), such as "
        "`fa6-solid:arrow-down`, or "
        "[superqt.fonticon](https://pyapp-kit.github.io/superqt/utilities/fonticon/)"
        " keys, such as `fa6s.arrow_down`",
    )

    @classmethod
    def _validate(cls, v: Any) -> "Icon":
        """Validate icon."""
        # if a single string is passed, use it for both light and dark.
        if isinstance(v, Icon):
            return v
        if isinstance(v, str):
            v = {"dark": v, "light": v}
        return cls(**v)

    # for v2
    @model_validator(mode="before")
    @classmethod
    def _model_val(cls, v: dict) -> dict:
        if isinstance(v, str):
            v = {"dark": v, "light": v}
        return v


class IconDict(TypedDict):
    """Icon dictionary."""

    dark: Optional[str]
    light: Optional[str]


IconOrDict = Union[Icon, IconDict]
